import json
import docker

DEFAULT_HOST = "unix://var/run/docker.sock"
# {"Id": "", "aka": {"RepoTags": [], "RepoDigests": []}}


class ErrorMessage(Exception):
    pass


def error_if(cond, msg):
    if cond:
        raise ErrorMessage(msg) from None


def prepare(docker_host, images):
    client = docker.DockerClient(docker_host, version="auto")
    images_idx = images_handle_index(client)

    results = []
    warnings = []
    for handle in images:
        hits = None
        for handle2 in (handle, handle + ":latest", "sha256:" + handle):
            if hits is None:
                hits = images_idx.get(handle2)
        error_if(not hits, "No such image: " + handle)
        error_if(len(hits) > 1, "Ambiguous image: " + handle)
        image_id = next(iter(hits))
        image_attrs = client.images.get(image_id).attrs
        res = {"imageId": image_id}
        if image_attrs.get("RepoTags"):
            res["akaRepoTags"] = image_attrs["RepoTags"]
        if image_attrs.get("RepoDigests"):
            res["akaRepoDigests"] = image_attrs["RepoDigests"]
        results.append(res)

    return (
        "\n".join(json.dumps(res, separators=(",", ":")) for res in results).encode() + b"\n"
    ), warnings


def verify(
    docker_host, sigbody, handle_to_verify=None, ignore_missing=False
):  # pylint: disable=R0912,R0914,R0915
    client = docker.DockerClient(docker_host, version="auto")
    local_images_index = images_handle_index(client)
    image_to_verify = None
    if handle_to_verify:
        image_to_verify = local_images_index.get(handle_to_verify)
        error_if(not image_to_verify, "No such local image: " + handle_to_verify)
        image_to_verify = client.images.get(next(iter(image_to_verify)))

    verified = []
    warnings = set()
    lines = [line for line in sigbody.split(b"\n") if line]
    for line in lines:
        try:
            sig_elt = json.loads(line.decode())
            assert isinstance(sig_elt.get("imageId"), str)
        except:
            error_if(True, "Invalid signature syntax")

        # First, check that if the signature includes tags, those tags don't point to a different
        # local image (exception: warning for :latest)
        signed_tags = []
        if "akaRepoTags" in sig_elt:
            assert isinstance(sig_elt["akaRepoTags"], list)
            signed_tags.extend(sig_elt["akaRepoTags"])
        if "akaRepoDigests" in sig_elt:
            assert isinstance(sig_elt["akaRepoDigests"], list)
            signed_tags.extend(sig_elt["akaRepoDigests"])
        for signed_handle in signed_tags:
            assert isinstance(signed_handle, str)
            error_if(
                len(local_images_index.get(signed_handle, set())) > 1,
                f"The local image tag '{signed_handle}' is ambiguous",
            )
            if (
                signed_handle in local_images_index
                and next(iter(local_images_index[signed_handle])) != sig_elt["imageId"]
            ):
                if signed_handle.endswith(":latest"):
                    warnings.add(
                        f"The local default image tag '{signed_handle}' refers to a different image ID than was signed; be sure to use the full ID or digest to get the correct image."
                    )
                else:
                    error_if(
                        True,
                        f"The local image tag '{signed_handle}' refers to a different image ID than was signed",
                    )

        if image_to_verify:
            if image_to_verify.id != sig_elt["imageId"]:
                continue
            local_image = image_to_verify
        else:
            local_image = local_images_index.get(sig_elt["imageId"])
            error_if(
                not (ignore_missing or local_image),
                "Signed image missing locally"
                + (
                    "; try --ignore-missing if OK for some but not all to be missing"
                    if len(lines) > 1
                    else ""
                ),
            )
            if local_image:
                assert len(local_image) == 1
                local_image = client.images.get(next(iter(local_image)))
            else:
                warnings.add("The transaction signs one or more images that are missing locally")
                continue

        local_tags = []
        if local_image.attrs.get("RepoTags"):
            local_tags.extend(local_image.attrs["RepoTags"])
        if local_image.attrs.get("RepoDigests"):
            local_tags.extend(local_image.attrs["RepoDigests"])
        common_tags = set(local_tags).intersection(set(signed_tags))

        aka_msg = ""
        if common_tags:
            aka_msg = ", aka: " + " ".join(sorted(common_tags))
        elif signed_tags and local_tags:
            warnings.add(
                f"Image ID = {local_image.id} was signed, but under different tag(s) than it's known by locally; double-check it's the intended image if selecting by tag."
            )
        verified.append("Verified image ID = " + local_image.id + aka_msg)

    error_if(not verified, "No image verified")
    return verified, warnings


def images_handle_index(client):
    # omnibus index of docker image IDs by ID, short ID, RepoTags, RepoDigests
    ans = {}
    for image in client.images.list():
        attrs = image.attrs
        assert image.id == attrs["Id"]
        ans.setdefault(image.id, set()).add(image.id)
        ans.setdefault(image.id[:12], set()).add(image.id)
        ans.setdefault(image.short_id, set()).add(image.id)
        for tag in attrs.get("RepoTags", []):
            ans.setdefault(tag, set()).add(image.id)
        for dig in attrs.get("RepoDigests", []):
            ans.setdefault(dig, set()).add(image.id)
    return ans



import sys
import json
import argparse
import platform
import subprocess
import shutil
from datetime import datetime, timedelta
import dateutil
import dateutil.parser
import dateutil.tz
import web3
from .verify import print_tsv, bail, yellow, color, ANSI


def prepare_sha256sum(files, sha256sum_exe, cwd=None, tee=False):
    "run sha256sum on files; prepare input body for signing transaction as bytes"
    # tee sha256sum stdout in realtime, to provide feedback whilst processing multiple large files
    proc = subprocess.Popen([sha256sum_exe] + files, stdout=subprocess.PIPE, cwd=cwd)
    sha256sum_stdout = []
    while True:
        line = proc.stdout.readline()
        if not line:
            break
        sha256sum_stdout.append(line)  # includes newline
        if tee:
            sys.stdout.buffer.write(line)
            sys.stdout.buffer.flush()
    proc.wait()
    sys.stdout.flush()
    if proc.returncode != 0:
        raise Exception("sha256sum failed")

    return b"".join(sha256sum_stdout)


def cli_subparser(subparsers):
    parser = subparsers.add_parser(
        "prepare",
        help="prepare data for signature",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("FILE", nargs="+", help="filenames or other object identifiers")
    parser.add_argument(
        "--git",
        action="store_true",
        help="identifiers are git commits or tags in the current repository",
    )
    parser.add_argument(
        "--docker",
        action="store_true",
        help="identifiers pertain to docker images from the local dockerd",
    )
    parser.add_argument(
        "--stake",
        metavar="0.1",
        dest="stake_ad",
        type=float,
        help="ETH stake to advertise/advise",
    )
    parser.add_argument(
        "--expire",
        metavar=f"'{datetime.utcnow()}Z'",
        help="declare signature expires at ISO 8601 date & time",
    )
    parser.add_argument(
        "--expire-days",
        metavar="N",
        type=int,
        help="declare signature expires N days from now",
    )
    parser.add_argument(
        "--chdir", "-C", metavar="DIR", type=str, help="change working directory to DIR"
    )
    return parser


def cli(args):  # pylint: disable=R0912,R0915
    if args.expire and args.expire_days:
        bail("set at most one of --expire-days and --expire")
    if args.docker and args.git:
        bail("set at most one of --git and --docker")
    if args.stake_ad is None:
        print(
            yellow(
                "[WARN] Are you sure you don't want to set the advisory --stake? This will leave it to verifiers to apply a default threshold.",
            )
        )

    expire_utc = None
    if args.expire:
        expire_utc = dateutil.parser.isoparse(args.expire).astimezone(dateutil.tz.tzutc())
    if args.expire_days is not None:
        expire_utc = datetime.utcnow() + timedelta(days=args.expire_days)
    if expire_utc is not None:
        expire_utc = expire_utc.replace(tzinfo=None)

    header = {"stakesign": "sha256sum"}
    if args.git:
        header["stakesign"] = "git"
    if args.docker:
        header["stakesign"] = "docker"
    if isinstance(expire_utc, datetime):
        header["expire"] = f"{expire_utc}Z"
    if isinstance(args.stake_ad, float):
        header["stakeAd"] = {"ETH": args.stake_ad}
    header = json.dumps(header, separators=(",", ":")) + "\n"

    if args.git:
        from .git import repository, prepare, ErrorMessage  # pylint: disable=C0415

        repo_dir, repo = repository(args.chdir)
        print_tsv("Trusting git repo:", repo_dir)

        try:
            body, warnings = prepare(repo, args.FILE)
        except ErrorMessage as err:
            bail(err.args[0])
        for warnmsg in warnings:
            print(yellow("[WARN] " + warnmsg))
        sys.stdout.flush()
        sys.stdout.buffer.write(header.encode())  # for payload preview
        sys.stdout.buffer.write(body)
    elif args.docker:
        from .docker import DEFAULT_HOST, prepare, ErrorMessage  # pylint: disable=C0415

        print_tsv("Trusting dockerd:", DEFAULT_HOST)
        try:
            body, warnings = prepare(DEFAULT_HOST, args.FILE)
        except ErrorMessage as err:
            bail(err.args[0])
        for warnmsg in warnings:
            print(yellow("[WARN] " + warnmsg))
        sys.stdout.flush()
        sys.stdout.buffer.write(header.encode())  # for payload preview
        sys.stdout.buffer.write(body)
    else:  # default sha256sum mode
        sha256sum_exe = shutil.which("sha256sum")
        if not sha256sum_exe:
            msg = "`sha256sum` utility unavailable; ensure coreutils is installed and PATH is configured"
            if platform.system() == "Darwin":
                msg += "\nOn macOS try: brew install coreutils"
            bail(msg)
        print_tsv("Trusting local exe:", sha256sum_exe)
        print()

        sys.stdout.write(header)  # for payload preview
        try:
            body = prepare_sha256sum(args.FILE, sha256sum_exe, cwd=args.chdir, tee=True)
        except:
            bail("`sha256sum` utility failed")

    print("\n-- Transaction input data for signing (one long line):\n")

    print(color(web3.Web3.toHex(header.encode() + body), ANSI.BOLD))
    print()



import json
import os
import sys
import argparse
import subprocess
import shutil
import tempfile
import math
from datetime import datetime, timedelta
import dateutil
import dateutil.tz
import dateutil.parser
import web3
from web3.datastructures import AttributeDict


DEFAULT_STAKE_FLOOR_ETH = 0.1


def get_sig(w3, txid):
    "Query blockchain for signature transaction details"
    tx = w3.eth.getTransaction(txid)
    txr = w3.eth.getTransactionReceipt(txid)
    block_num = tx.blockNumber
    if not block_num:
        raise web3.exceptions.TransactionNotFound("transaction pending (no block number yet)")
    blk = w3.eth.getBlock(block_num)

    signer = txr["from"]
    assert tx["from"] == signer

    return AttributeDict(
        {
            "id": txid,
            "timestamp": datetime.utcfromtimestamp(blk.timestamp),
            "block": tx.blockNumber,
            "signer": signer,
            "input": tx.input,
        }
    )


def decode_sig_input(w3, sig):
    "Decode signature input data to header dict & body bytes"
    assert isinstance(sig.input, str) and sig.input.startswith("0x")
    buf = w3.toBytes(hexstr=sig.input)
    pos = buf.find(b"\n")
    pos = pos if pos >= 0 else len(pos)
    try:
        hdr = json.loads(buf[:pos])
        assert isinstance(hdr, dict)
        assert "stakesign" in hdr and isinstance(hdr["stakesign"], str)
    except:
        raise ValueError(
            "Transaction input isn't consistent with stakesign format; check transaction ID"
        ) from None
    bod = buf[(pos + 1) :]
    return (hdr, bod)


def check_sig_expire(header, utcnow):
    assert isinstance(utcnow, datetime)
    expire = None
    if "expire" in header:
        try:
            assert isinstance(header["expire"], str)
            expire = (
                dateutil.parser.isoparse(header["expire"])
                .astimezone(dateutil.tz.tzutc())
                .replace(tzinfo=None)
            )
        except:
            raise ValueError(
                "Transaction header.expire has invalid value (expected ISO 8601)"
            ) from None
    return AttributeDict(
        {
            "unexpired": expire and expire > utcnow,
            "expire_utc": expire,
            "now_utc": utcnow,
        }
    )


def check_sig_stake(w3, sig, header, stake_floor_wei, ignore_ad=False):
    "Check whether the signing address has sufficient current ETH balance"
    assert isinstance(header, dict)
    assert isinstance(stake_floor_wei, int)

    signer_wei = w3.eth.getBalance(sig.signer)
    assert isinstance(signer_wei, int)

    required_wei = stake_floor_wei
    required_wei_source = "--stake"
    if not ignore_ad and "stakeAd" in header:
        stake_ad = header["stakeAd"]
        if not isinstance(stake_ad, dict):
            raise ValueError("Transaction header.stakeAd has invalid value")
        if "ETH" in stake_ad:
            stake_ad = stake_ad["ETH"]
            if not (isinstance(stake_ad, float) and math.isfinite(stake_ad)):
                raise ValueError("Transaction header.stakeAd has invalid ETH value")
            stake_ad = w3.toWei(stake_ad, "ether")
            if stake_ad > required_wei:
                required_wei = stake_ad
                required_wei_source = "stakeAd"
        else:
            print(color("[WARN] Transaction header.stakeAd doesn't specify ETH value", ANSI.BHYEL))

    return AttributeDict(
        {
            "enough": signer_wei >= required_wei,
            "signer_wei": signer_wei,
            "required_wei": required_wei,
            "required_wei_source": required_wei_source,
        }
    )


def verify_sha256sum(header, body, exe, ignore_missing=False, no_strict=False, cwd=None):
    "run given sha256sum executable to verify signature body"
    assert header["stakesign"] == "sha256sum"
    assert isinstance(body, bytes)

    cmd = [exe, "--check"]
    if not no_strict:
        cmd.append("--strict")
    if ignore_missing:
        cmd.append("--ignore-missing")

    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(body)
        tmp.flush()
        cmd.append(tmp.name)
        res = subprocess.run(cmd, check=False, cwd=cwd)

    return res.returncode == 0


def cli_subparser(subparsers):
    parser = subparsers.add_parser(
        "verify",
        help="verify existing signature(s)",
        description="The transaction is, by default, expected to sign file(s) in the current working directory.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("signature", help="signature Transaction ID (0x...)")
    parser.add_argument(
        "--stake",
        metavar="0.1",
        dest="stake_floor_eth",
        type=float,
        default=DEFAULT_STAKE_FLOOR_ETH,
        help="minimum acceptable current ETH balance for signer address",
    )
    parser.add_argument(
        "--ignore-ad",
        action="store_true",
        help="use --stake value even if less than signature's stakeAd",
    )
    parser.add_argument(
        "--ignore-missing",
        action="store_true",
        help="if signature covers multiple files/objects, only require at least one to be present",
    )
    parser.add_argument(
        "--no-strict",
        action="store_true",
        help="do not pass --strict to sha256sum (use only if yours doesn't provide this option)",
    )
    parser.add_argument(
        "--expired-ok",
        action="store_true",
        help="proceed even if signature's stated expiration date has passed",
    )
    parser.add_argument(
        "--git",
        dest="git_revision",
        metavar="HEAD",
        help="expect signature of git commits/tags & specify the local revision to verify (default HEAD)",
    )
    parser.add_argument(
        "--docker",
        dest="docker_handle",
        metavar="TAG",
        help="expect signature of docker image(s) and specify the local image tag/digest/ID to verify",
    )
    parser.add_argument(
        "--file", dest="files_only", action="store_true", help="expect signature of file(s)"
    )
    parser.add_argument(
        "--chdir", "-C", metavar="DIR", type=str, help="change working directory to DIR"
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="display transaction input payload after success (caution: prints data from blockchain to terminal)",
    )
    return parser


def cli(args):  # pylint: disable=R0912,R0914,R0915
    provider_msg = "(from environment WEB3_PROVIDER_URI)"
    if "WEB3_PROVIDER_URI" not in os.environ:
        os.environ["WEB3_PROVIDER_URI"] = "https://cloudflare-eth.com"
        provider_msg = "(to override, set environment WEB3_PROVIDER_URI)"

    print("\t".join(("Trusting ETH gateway:", os.environ["WEB3_PROVIDER_URI"], provider_msg)))
    from web3.auto import w3  # pylint: disable=C0415

    # get transaction info
    if not args.signature.startswith("0x"):
        bail("Transaction ID should start with 0x")
    try:
        sig = get_sig(w3, args.signature)
    except web3.exceptions.TransactionNotFound as err:
        bail(
            "Transaction not found on Ethereum network; check transaction ID, or try later or through another gateway: "
            + str(err)
        )

    utcnow = datetime.utcnow().replace(tzinfo=None)
    sig_age = utcnow - sig.timestamp
    print_tsv("         Transaction:", sig.id)
    print_tsv("    Signer's address:", sig.signer)
    print_tsv(
        " Signature timestamp:",
        f"{sig.timestamp}Z",
        yellow(f"({sig_age} ago)", sig_age < timedelta(days=3)),
    )

    # decode signature, check expiration date
    header, body = decode_sig_input(w3, sig)
    exinfo = check_sig_expire(header, utcnow)
    if exinfo.expire_utc is not None:
        print_tsv(
            "Signature expiration:",
            f"{exinfo.expire_utc}Z",
            f"{color('🗹', ANSI.BHGRN) if exinfo.unexpired else color('✗', ANSI.BHRED)}",
        )
        if not (args.expired_ok or exinfo.unexpired):
            bail("Signature's stated expiration date has passed")

    # check stake
    vs = check_sig_stake(
        w3,
        sig,
        header,
        w3.toWei(args.stake_floor_eth, "ether"),
        ignore_ad=args.ignore_ad,
    )
    print_tsv(
        "Signer's balance now:",
        f"{w3.fromWei(vs.signer_wei, 'ether')}",
        f"{'≥' if vs.enough else '<'} {w3.fromWei(vs.required_wei, 'ether')} ETH from {vs.required_wei_source}",
        f"{color('🗹', ANSI.BHGRN) if vs.enough else color('✗', ANSI.BHRED)}",
    )
    if not vs.enough:
        msg = "Signer's address holds insufficient ETH balance, possibly indicating revocation or compromise!"
        if vs.required_wei_source == "--stake":
            msg += f"\n        If you're certain this address is trustworthy, rerun with --stake {w3.fromWei(vs.signer_wei, 'ether')}"
        bail(msg)

    # verify, per mode
    warnings = []
    mode = header["stakesign"]
    if mode == "sha256sum":
        if args.git_revision:
            bail("Signature applies to files, not git")
        if args.docker_handle:
            bail("Signature applies to files, not docker")
        sha256sum_exe = shutil.which("sha256sum")
        if not sha256sum_exe:
            bail(
                "`sha256sum` utility unavailable; ensure coreutils is installed and PATH is configured"
            )
        print_tsv("  Trusting local exe:", sha256sum_exe)
        print()
        if not verify_sha256sum(
            header,
            body,
            sha256sum_exe,
            ignore_missing=args.ignore_missing,
            no_strict=args.no_strict,
            cwd=args.chdir,
        ):
            bail("sha256sum verification failed!")
    elif mode == "git":
        if args.files_only:
            bail("Signature applies to git, not files")
        if args.docker_handle:
            bail("Signature applies to git, not docker")
        from .git import repository, verify, ErrorMessage  # pylint: disable=C0415

        try:
            repo_dir, repo = repository(args.chdir)
        except:
            bail(
                "Signature pertains to git commit, but current working directory isn't a git repository"
            )
        revision = args.git_revision if args.git_revision else "HEAD"
        print_tsv("Local git repository:", repo_dir)
        print_tsv("  Local git revision:", revision)
        try:
            msg, warnings = verify(repo, revision, body, ignore_missing=args.ignore_missing)
        except ErrorMessage as err:
            bail(err.args[0])
        print()
        print(msg)
        print()
    elif mode == "docker":
        if args.files_only:
            bail("Signature applies to docker, not files")
        if args.git_revision:
            bail("Signature applies to docker, not git")
        from .docker import DEFAULT_HOST, verify, ErrorMessage  # pylint: disable=C0415

        print_tsv("    Trusting dockerd:", DEFAULT_HOST)
        try:
            verifications, warnings = verify(
                DEFAULT_HOST, body, args.docker_handle, args.ignore_missing
            )
        except ErrorMessage as err:
            bail(err.args[0])
        print()
        for msg in verifications:
            print(msg)
        print()
    else:
        bail(
            "Signing mode not one of {sha256sum, git, docker}; a newer version of this utility might support the necessary mode."
        )

    for warnmsg in warnings:
        print(yellow("[WARN] " + warnmsg))
    if math.fabs(args.stake_floor_eth - DEFAULT_STAKE_FLOOR_ETH) < (DEFAULT_STAKE_FLOOR_ETH / 1000):
        print(
            yellow(
                f"[WARN] Ensure the signer's current {w3.fromWei(vs.signer_wei, 'ether')} ETH stake evinces their ongoing interest in securing it.\n"
                + f"       (Set --stake above the default {DEFAULT_STAKE_FLOOR_ETH} ETH minimum to clear this warning.)",
            )
        )
        warnings = True
    print_tsv(
        color("🗹", ANSI.BHGRN),
        color("Success (with warnings)" if warnings else "Success", ANSI.BOLD),
    )

    if args.verbose:
        print()
        print(w3.toBytes(hexstr=sig.input).decode("utf-8").rstrip("\n"))


def print_tsv(*args, **kwargs):
    print("\t".join(str(arg) for arg in args), **kwargs)


def bail(msg):
    msg = "[ERROR] " + msg
    if sys.stderr.isatty() and "NO_COLOR" not in os.environ:
        print(ANSI.BHRED + msg + ANSI.RESET, file=sys.stderr)
    else:
        print(msg, file=sys.stderr)
    sys.exit(1)


def color(msg, col):
    if sys.stdout.isatty() and "NO_COLOR" not in os.environ:
        return col + msg + ANSI.RESET
    return msg


def yellow(msg, only_if=True):
    return color(msg, ANSI.BHYEL) if only_if else msg


class ANSI:
    # https://gist.github.com/RabaDabaDoba/145049536f815903c79944599c6f952a
    # https://espterm.github.io/docs/VT100%20escape%20codes.html
    RESET = "\x1b[0m"
    BHRED = "\x1b[1;91m"
    BHYEL = "\x1b[1;93m"
    BHGRN = "\x1b[1;92m"
    BOLD = "\x1b[1m"



import json
from pathlib import Path
from pygit2 import Repository, Commit, Reference, Tag  # pylint: disable=E0611


class ErrorMessage(Exception):
    pass


def error_if(cond, msg):
    if cond:
        raise ErrorMessage(msg) from None


def repository(cwd=None):
    # find containing git repo
    cwd = (Path(cwd) if cwd else Path.cwd()).resolve()
    while not (cwd / ".git").is_dir():
        cwd = (cwd / "..").resolve()
        error_if(cwd == Path(cwd.anchor), "Not in git repository")
    return str(cwd), Repository(cwd / ".git")


def prepare(repo, revisions):
    results = []
    warnings = []
    all_sha256 = True
    for revision in revisions:
        try:
            obj, ref = repo.revparse_ext(revision)
        except KeyError:
            error_if(True, f"Failed to `git rev-parse {revision}`")
        assert isinstance(obj, (Commit, Tag))
        assert not ref or isinstance(ref, Reference)
        result = {}
        if isinstance(obj, Commit):
            result["commit"] = obj.hex
            all_sha256 = all_sha256 and len(obj.hex) == 64
            if ref and ref.name.startswith("refs/tags/"):  # lightweight tag
                result["tag"] = ref.name[10:]
        elif isinstance(obj, Tag):  # annotated tag
            result["commit"] = obj.target.hex
            result["tag"] = obj.name
            result["tagObject"] = obj.hex
            all_sha256 = all_sha256 and len(obj.target.hex) == 64 and len(obj.hex) == 64
        else:
            assert False
        results.append(result)

    head_commit = repo.revparse_ext("HEAD")[0].hex
    if head_commit not in (res["commit"] for res in results):
        warnings.append(
            f"The revisions to sign don't include the current working tree HEAD = {head_commit}"
        )
    elif repo.status():
        warnings.append(
            "Working tree is dirty; signature will apply to clean commit HEAD = " + head_commit
        )

    if not all_sha256:
        warnings.append(
            "Preparing signature for git SHA-1 digest; review git SHA-1 security risks and consider adopting git SHA-256 mode"
        )
    return (
        "\n".join(json.dumps(res, separators=(",", ":")) for res in results).encode() + b"\n"
    ), warnings


def verify(repo, revision, sigbody, ignore_missing=False):  # pylint: disable=R0912,R0915
    """
    Verify that signature body includes a valid signature of revision.
    """
    # resolve revision to commit hash
    try:
        obj_to_verify, _ = repo.revparse_ext(revision)
    except KeyError:
        error_if(True, f"Failed to `git rev-parse {revision}`")
    if isinstance(obj_to_verify, Commit):
        commit_to_verify = obj_to_verify.hex
    elif isinstance(obj_to_verify, Tag):
        commit_to_verify = obj_to_verify.target.hex
    else:
        assert False

    # check status of current checkout
    warnings = set()
    head_commit = repo.revparse_ext("HEAD")[0].hex
    if head_commit != commit_to_verify:
        warnings.add(
            f"Verified revision{revision} = {commit_to_verify} is not the working tree HEAD = {head_commit}"
        )
    elif repo.status():
        warnings.add(
            "Working tree is dirty; signature applies to clean commit HEAD = " + head_commit
        )

    verified = None
    all_sha256 = len(commit_to_verify) == 64

    # Look for signature of commit_to_verify
    # Warning about warning messages: sigbody comes off the blockchain, so we shouldn't include
    # anything from it in warning messages without validation (in case it is malicious)
    lines = [line for line in sigbody.split(b"\n") if line]
    for line in lines:  # pylint: disable=R1702
        try:
            sig_elt = json.loads(line.decode())
            assert isinstance(sig_elt.get("commit"), str)
        except:
            error_if(True, "Invalid signature syntax")
        if not repo.get(sig_elt["commit"]):
            error_if(
                not ignore_missing,
                (
                    "Signed commit missing from local repository"
                    if len(lines) <= 1
                    else "Signed commit(s) missing from local repository; try --ignore-missing if OK for some but not all to be present"
                ),
            )
            warnings.add("One or more signed commit(s) missing from local repository")
        # If the signature includes tags, make sure they don't refer to commits other than the
        # signed ones. There are several cases to deal with here as the signed & local tags could
        # each be either lightweight or annotated.
        local_tag = None
        if "tag" in sig_elt:
            assert isinstance(sig_elt["tag"], str)
            assert "tagObject" not in sig_elt or isinstance(sig_elt["tagObject"], str)
            try:
                local_tag = repo.revparse_ext(sig_elt["tag"])
            except KeyError:
                error_if(
                    not ignore_missing,
                    "Signed tag(s) missing from local repository; try --ignore-missing if this is OK",
                )
                warnings.add("One or more signed tag(s) missing from local repository")
                local_tag = (None, None)
            if isinstance(local_tag[0], Commit):  # local lightweight tag
                assert isinstance(local_tag[1], Reference)
                error_if(
                    not local_tag[1].name.startswith("refs/tags/"),
                    "The signed tag refers locally to something else: " + local_tag[1].name,
                )
                error_if(
                    local_tag[0].hex != sig_elt["commit"],
                    f"The local tag '{local_tag[1].shorthand}' refers to a different commit than the signed tag",
                )
                if "tagObject" in sig_elt:
                    warnings.add(
                        f"The local tag '{local_tag[1].shorthand}' is lightweight, while the signed tag was annotated"
                    )
            elif isinstance(local_tag[0], Tag):  # local annotated tag
                if "tagObject" in sig_elt:
                    error_if(
                        sig_elt["tagObject"] != local_tag[0].hex,
                        f"The local tag '{local_tag[0].name}' = {local_tag[0].hex} differs from the signed tag in annotations (although they share the same name and commit reference)",
                    )
                else:
                    error_if(
                        local_tag[0].target.hex != sig_elt["commit"],
                        f"The local annotated tag '{local_tag[0].name}' = {local_tag[0].hex} refers to a different commit than the signed tag",
                    )
                    warnings.add(
                        f"The local tag '{local_tag[1].shorthand}' is annotated, while the signed tag was lightweight"
                    )
                all_sha256 = all_sha256 and len(local_tag[0].hex) == 64
            elif local_tag[0] is not None:
                assert False
        # At last...check whether sig_elt signs the desired commit
        if sig_elt["commit"] == commit_to_verify:
            if local_tag and local_tag[1]:
                verified = f"Verified: local revision {revision} = signed tag {local_tag[1].shorthand} (commit {commit_to_verify})"
            elif verified is None:
                verified = f"Verified: local revision {revision} = signed commit {commit_to_verify}"

    error_if(not verified, f"Signature doesn't apply to {revision} ({commit_to_verify})")
    if not all_sha256:
        warnings.add(
            "Signature pertains to git SHA-1 digest(s); review git SHA-1 security risks and consider adopting git SHA-256 mode"
        )
    return verified, warnings



import os
import sys
from argparse import ArgumentParser, Action
import importlib_metadata
from . import verify, prepare


def main():
    parser = ArgumentParser("stakesign")
    parser.add_argument(
        "--version",
        nargs=0,
        action=PipVersionAction,
        help="show package version information",
    )

    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.dest = "command"
    verify.cli_subparser(subparsers)
    prepare.cli_subparser(subparsers)

    replace_COLUMNS = os.environ.get("COLUMNS", None)
    os.environ["COLUMNS"] = "120"  # make help descriptions wider
    args = parser.parse_args(sys.argv[1:])
    if replace_COLUMNS is not None:
        os.environ["COLUMNS"] = replace_COLUMNS
    else:
        del os.environ["COLUMNS"]

    stakesign_version = "(version unknown)"
    try:
        stakesign_version = "v" + importlib_metadata.version("stakesign")
    except importlib_metadata.PackageNotFoundError:
        pass
    print(
        f"[NOTICE] stakesign {stakesign_version} is a prototype; don't trust in high-risk environments"
    )

    if args.command == "verify":
        verify.cli(args)
    elif args.command == "prepare":
        prepare.cli(args)
    else:
        assert False


class PipVersionAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        from web3.auto import w3  # pylint: disable=C0415

        print(f"web3 v{w3.api}")
        try:
            print(f"stakesign v{importlib_metadata.version('stakesign')}")
        except importlib_metadata.PackageNotFoundError:
            print("stakesign version unknown")
        sys.exit(0)



#!/usr/bin/env python3
from os import path
from setuptools import setup, find_packages
from version import get_version

HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, "README.md")) as f:
    README = f.read()

setup(
    name="stakesign",
    version=get_version(),
    url="https://github.com/mlin/stakesign",
    description="Sign files via blockchain + put your money where your mouth is",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Mike Lin",
    author_email="dna@mlin.net",
    license="MIT",
    packages=find_packages(),
    entry_points={"console_scripts": ["stakesign = stakesign:main"]},
    python_requires=">=3.6",
)
